from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Generator, overload

from kontainer.container.maybe import Maybe
from kontainer.core.types import Container
from kontainer.utils.generator import unwrap_generator

if TYPE_CHECKING:
    from typing_extensions import ParamSpec, TypeVar

    ValueT = TypeVar("ValueT", infer_variance=True)
    ParamT = ParamSpec("ParamT")

__all__ = ["optional"]


@overload
def optional(
    func: Callable[ParamT, Generator[Any, Any, None]]
) -> Callable[ParamT, Maybe[Any]]: ...


@overload
def optional(
    func: Callable[ParamT, Generator[Any, Any, Maybe[ValueT]]]
) -> Callable[ParamT, Maybe[ValueT]]: ...


@overload
def optional(
    func: Callable[ParamT, Generator[Any, Any, ValueT | None]]
) -> Callable[ParamT, Maybe[ValueT]]: ...


@overload
def optional(
    func: Callable[ParamT, Generator[Any, Any, ValueT]]
) -> Callable[ParamT, Maybe[ValueT]]: ...


@overload
def optional(
    func: Callable[ParamT, Awaitable[Maybe[ValueT]]]
) -> Callable[ParamT, Maybe[ValueT]]: ...


@overload
def optional(
    func: Callable[ParamT, Awaitable[ValueT]]
) -> Callable[ParamT, Maybe[ValueT]]: ...


@overload
def optional(func: Callable[ParamT, None]) -> Callable[ParamT, Maybe[Any]]: ...


@overload
def optional(
    func: Callable[ParamT, Maybe[ValueT]]
) -> Callable[ParamT, Maybe[ValueT]]: ...


@overload
def optional(
    func: Callable[ParamT, ValueT | None]
) -> Callable[ParamT, Maybe[ValueT]]: ...


@overload
def optional(func: Callable[ParamT, ValueT]) -> Callable[ParamT, Maybe[ValueT]]: ...


def optional(
    func: Callable[ParamT, Generator[Any, Any, None]]
    | Callable[ParamT, Generator[Any, Any, Maybe[ValueT]]]
    | Callable[ParamT, Generator[Any, Any, ValueT | None]]
    | Callable[ParamT, Generator[Any, Any, ValueT]]
    | Callable[ParamT, Awaitable[Maybe[ValueT]]]
    | Callable[ParamT, Awaitable[ValueT]]
    | Callable[ParamT, None]
    | Callable[ParamT, Maybe[ValueT]]
    | Callable[ParamT, ValueT | None]
    | Callable[ParamT, ValueT],
) -> Callable[ParamT, Maybe[Any]] | Callable[ParamT, Maybe[ValueT]]:
    @wraps(func)
    def inner(*args: ParamT.args, **kwargs: ParamT.kwargs) -> Maybe[Any]:
        result = func(*args, **kwargs)
        if not isinstance(result, Container) and isinstance(result, Awaitable):
            result = result.__await__()
        if isinstance(result, Generator):
            result = unwrap_generator(result)
        if isinstance(result, Maybe):
            return Maybe(result._value)  # noqa: SLF001
        return Maybe(result)

    return inner



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Generator, Generic, Iterable

from typing_extensions import Self, TypeVar

from kontainer.utils.generator import create_generator

ValueT = TypeVar("ValueT", infer_variance=True)
OtherT = TypeVar("OtherT", infer_variance=True)
if TYPE_CHECKING:
    AnotherT = TypeVar("AnotherT", infer_variance=True)
    ElementT = TypeVar("ElementT", infer_variance=True)

__all__ = ["Container"]


class Container(Iterable[ValueT], Awaitable[ValueT], Generic[ValueT, OtherT], ABC):
    r"""Container

    map: (x, y) -> (X, y)
    bind: (x, y) -> (X, Y)
    """

    __slots__ = ("_value",)

    @abstractmethod
    def __init__(self, value: ValueT) -> None: ...

    @abstractmethod
    def __repr__(self) -> str: ...

    @abstractmethod
    def __str__(self) -> str: ...

    @abstractmethod
    def __eq__(self, other: object) -> bool: ...

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    @abstractmethod
    def __hash__(self) -> int: ...

    def __iter__(self) -> Generator[Any, Any, ValueT]:
        return create_generator(self.unwrap())

    def __await__(self) -> Generator[Any, Any, ValueT]:
        return iter(self)

    def __copy__(self) -> Self:
        return self.copy()

    def __deepcopy__(self, memo: dict[Any, Any] | None = None) -> Self:
        new = self.deepcopy()
        if memo is None:
            return new
        memo[id(new)] = new
        return new

    @abstractmethod
    def map_value(
        self, func: Callable[[ValueT], AnotherT]
    ) -> Container[AnotherT, OtherT]:
        """_summary_

        value -> new value -> (new value, other)

        Args:
            func: _description_

        Returns:
            _description_
        """

    @abstractmethod
    def map_values(
        self, value: ElementT, func: Callable[[ValueT, ElementT], AnotherT]
    ) -> Container[AnotherT, OtherT]:
        """_summary_

        value, arg -> new value -> (new value, other)

        Args:
            value: _description_
            func: _description_

        Returns:
            _description_
        """

    @abstractmethod
    def map_container(
        self,
        value: Container[ElementT, Any],
        func: Callable[[ValueT, ElementT], AnotherT],
    ) -> Container[AnotherT, OtherT]:
        """_summary_

        value, container -> new value -> (new value, other)

        Args:
            value: _description_
            func: _description_

        Returns:
            _description_
        """

    @abstractmethod
    def bind_value(
        self, func: Callable[[ValueT], Container[AnotherT, OtherT]]
    ) -> Container[AnotherT, OtherT]:
        """_summary_

        value -> new value, new other -> (new value, new other)

        Args:
            func: _description_

        Returns:
            _description_
        """

    @abstractmethod
    def bind_values(
        self,
        value: ElementT,
        func: Callable[[ValueT, ElementT], Container[AnotherT, OtherT]],
    ) -> Container[AnotherT, OtherT]:
        """_summary_

        value, arg -> new value, new other -> (new value, new other)

        Args:
            value: _description_
            func: _description_

        Returns:
            _description_
        """

    @abstractmethod
    def bind_container(
        self,
        value: Container[ElementT, Any],
        func: Callable[[ValueT, ElementT], Container[AnotherT, OtherT]],
    ) -> Container[AnotherT, OtherT]:
        """_summary_

        value, container -> new value, new other -> (new value, new other)

        Args:
            value: _description_
            func: _description_

        Returns:
            _description_
        """

    @abstractmethod
    def switch(self) -> Container[OtherT, ValueT]: ...

    @abstractmethod
    def default(self, value: AnotherT) -> ValueT | AnotherT: ...

    @abstractmethod
    def map_default(self, func: Callable[[], AnotherT]) -> ValueT | AnotherT: ...

    @abstractmethod
    def unwrap(self) -> ValueT: ...

    @property
    @abstractmethod
    def is_positive(self) -> bool: ...

    @property
    def is_negative(self) -> bool:
        return not self.is_positive

    @abstractmethod
    def copy(self) -> Self: ...

    @abstractmethod
    def deepcopy(self) -> Self: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Iterable, Literal, overload

if TYPE_CHECKING:
    from typing_extensions import TypeVar

    from kontainer.container import Maybe, Result
    from kontainer.core.types import Container

    ValueT = TypeVar("ValueT", infer_variance=True)
    OtherT = TypeVar("OtherT", infer_variance=True)
    AnotherT = TypeVar("AnotherT", infer_variance=True)

__all__ = ["map_elements", "bind_elements"]


@overload
def map_elements(
    values: Iterable[Maybe[ValueT]], func: Callable[[ValueT], OtherT], /
) -> Iterable[Maybe[OtherT]]: ...


@overload
def map_elements(
    values: Iterable[Maybe[ValueT]],
    func: Callable[[ValueT], OtherT],
    /,
    *,
    lazy: Literal[False] = ...,
) -> tuple[Maybe[OtherT], ...]: ...


@overload
def map_elements(
    values: Iterable[Maybe[ValueT]],
    func: Callable[[ValueT], OtherT],
    /,
    *,
    lazy: Literal[True],
) -> Iterable[Maybe[OtherT]]: ...


@overload
def map_elements(
    values: Iterable[Maybe[ValueT]],
    func: Callable[[ValueT], OtherT],
    /,
    *,
    lazy: bool = ...,
) -> Iterable[Maybe[OtherT]] | tuple[Maybe[OtherT], ...]: ...


@overload
def map_elements(
    values: Iterable[Result[ValueT, Any]], func: Callable[[ValueT], OtherT], /
) -> Iterable[Result[OtherT, Exception]]: ...


@overload
def map_elements(
    values: Iterable[Result[ValueT, Any]],
    func: Callable[[ValueT], OtherT],
    /,
    *,
    lazy: Literal[False],
) -> tuple[Result[OtherT, Exception], ...]: ...


@overload
def map_elements(
    values: Iterable[Result[ValueT, Any]],
    func: Callable[[ValueT], OtherT],
    /,
    *,
    lazy: Literal[True],
) -> Iterable[Result[OtherT, Exception]]: ...


@overload
def map_elements(
    values: Iterable[Result[ValueT, Any]],
    func: Callable[[ValueT], OtherT],
    /,
    *,
    lazy: bool = ...,
) -> Iterable[Result[OtherT, Exception]] | tuple[Result[OtherT, Exception], ...]: ...


@overload
def map_elements(
    values: Iterable[Container[ValueT, Any]], func: Callable[[ValueT], OtherT], /
) -> Iterable[Container[OtherT, Any]]: ...


@overload
def map_elements(
    values: Iterable[Container[ValueT, Any]],
    func: Callable[[ValueT], OtherT],
    /,
    *,
    lazy: Literal[False],
) -> tuple[Container[OtherT, Any], ...]: ...


@overload
def map_elements(
    values: Iterable[Container[ValueT, Any]],
    func: Callable[[ValueT], OtherT],
    /,
    *,
    lazy: Literal[True],
) -> Iterable[Container[OtherT, Any]]: ...


@overload
def map_elements(
    values: Iterable[Container[ValueT, Any]],
    func: Callable[[ValueT], OtherT],
    /,
    *,
    lazy: bool = ...,
) -> Iterable[Container[OtherT, Any]] | tuple[Container[OtherT, Any], ...]: ...


def map_elements(
    values: Iterable[Container[ValueT, Any]],
    func: Callable[[ValueT], OtherT],
    /,
    *,
    lazy: bool = True,
) -> Iterable[Container[OtherT, Any]] | tuple[Container[OtherT, Any], ...]:
    result = (x.map_value(func) for x in values)
    if lazy:
        return result
    return tuple(result)


@overload
def bind_elements(
    values: Iterable[Maybe[ValueT]], func: Callable[[ValueT], Maybe[OtherT]], /
) -> Iterable[Maybe[OtherT]]: ...


@overload
def bind_elements(
    values: Iterable[Maybe[ValueT]],
    func: Callable[[ValueT], Maybe[OtherT]],
    /,
    *,
    lazy: Literal[False] = ...,
) -> tuple[Maybe[OtherT], ...]: ...


@overload
def bind_elements(
    values: Iterable[Maybe[ValueT]],
    func: Callable[[ValueT], Maybe[OtherT]],
    /,
    *,
    lazy: Literal[True],
) -> Iterable[Maybe[OtherT]]: ...


@overload
def bind_elements(
    values: Iterable[Maybe[ValueT]],
    func: Callable[[ValueT], Maybe[OtherT]],
    /,
    *,
    lazy: bool = ...,
) -> Iterable[Maybe[OtherT]] | tuple[Maybe[OtherT], ...]: ...


@overload
def bind_elements(
    values: Iterable[Result[ValueT, Any]],
    func: Callable[[ValueT], Result[OtherT, Any]],
    /,
) -> Iterable[Result[OtherT, Exception]]: ...


@overload
def bind_elements(
    values: Iterable[Result[ValueT, Any]],
    func: Callable[[ValueT], Result[OtherT, Any]],
    /,
    *,
    lazy: Literal[False],
) -> tuple[Result[OtherT, Exception], ...]: ...


@overload
def bind_elements(
    values: Iterable[Result[ValueT, Any]],
    func: Callable[[ValueT], Result[OtherT, Any]],
    /,
    *,
    lazy: Literal[True],
) -> Iterable[Result[OtherT, Exception]]: ...


@overload
def bind_elements(
    values: Iterable[Result[ValueT, Any]],
    func: Callable[[ValueT], Result[OtherT, Any]],
    /,
    *,
    lazy: bool = ...,
) -> Iterable[Result[OtherT, Exception]] | tuple[Result[OtherT, Exception], ...]: ...


@overload
def bind_elements(
    values: Iterable[Container[ValueT, Any]],
    func: Callable[[ValueT], Container[OtherT, Any]],
    /,
) -> Iterable[Container[OtherT, Any]]: ...


@overload
def bind_elements(
    values: Iterable[Container[ValueT, Any]],
    func: Callable[[ValueT], Container[OtherT, Any]],
    /,
    *,
    lazy: Literal[False],
) -> tuple[Container[OtherT, Any], ...]: ...


@overload
def bind_elements(
    values: Iterable[Container[ValueT, Any]],
    func: Callable[[ValueT], Container[OtherT, Any]],
    /,
    *,
    lazy: Literal[True],
) -> Iterable[Container[OtherT, Any]]: ...


@overload
def bind_elements(
    values: Iterable[Container[ValueT, Any]],
    func: Callable[[ValueT], Container[OtherT, Any]],
    /,
    *,
    lazy: bool = ...,
) -> Iterable[Container[OtherT, Any]] | tuple[Container[OtherT, Any], ...]: ...


def bind_elements(
    values: Iterable[Container[ValueT, Any]],
    func: Callable[[ValueT], Container[OtherT, Any]],
    /,
    *,
    lazy: bool = True,
) -> Iterable[Container[OtherT, Any]] | tuple[Container[OtherT, Any], ...]:
    result = (x.bind_value(func) for x in values)
    if lazy:
        return result
    return tuple(result)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, overload

from kontainer.core.exception import KontainerTypeError

if TYPE_CHECKING:
    from typing_extensions import TypeVar

    from kontainer.container import Maybe, Result

    ValueT = TypeVar("ValueT", infer_variance=True)

__all__ = ["toggle"]


@overload
def toggle(container: Result[ValueT, Any]) -> Maybe[ValueT]: ...


@overload
def toggle(container: Maybe[ValueT]) -> Result[ValueT, None]: ...


@overload
def toggle(
    container: Maybe[ValueT] | Result[ValueT, Any]
) -> Result[ValueT, None] | Maybe[ValueT]: ...


def toggle(
    container: Maybe[ValueT] | Result[ValueT, Any]
) -> Result[ValueT, None] | Maybe[ValueT]:
    from kontainer.container.maybe import Null, Some
    from kontainer.container.result import Done, Error

    if isinstance(container, Some):
        return Done(container.unwrap())
    if isinstance(container, Null):
        return Error(None)
    if isinstance(container, Done):
        return Some(container.unwrap())
    if isinstance(container, Error):
        return Null(None)

    error_msg = f"invalid container type: {type(container).__name__!r}"
    raise KontainerTypeError(error_msg)



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Generator, Generic, overload

from typing_extensions import ParamSpec, TypeVar

from kontainer.container.result import Error, Result
from kontainer.core.types import Container
from kontainer.utils.generator import unwrap_generator

ErrorT = TypeVar("ErrorT", infer_variance=True, bound=Exception)
if TYPE_CHECKING:
    ValueT = TypeVar("ValueT", infer_variance=True)
    ErrorT2 = TypeVar("ErrorT2", infer_variance=True, bound=Exception)
    OtherValueT = TypeVar("OtherValueT", infer_variance=True)
    OtherErrorT = TypeVar("OtherErrorT", infer_variance=True, bound=Exception)
    OtherErrorT2 = TypeVar("OtherErrorT2", infer_variance=True, bound=Exception)
    ParamT = ParamSpec("ParamT")
    OtherParamT = ParamSpec("OtherParamT")


__all__ = ["catch"]


class _Catch(Generic[ErrorT]):
    __slots__ = ("_error_type",)

    def __init__(self, error_type: type[ErrorT]) -> None:
        self._error_type = error_type

    @overload
    def __call__(
        self, func: Callable[ParamT, Generator[Any, Any, Result[ValueT, ErrorT2]]]
    ) -> Callable[ParamT, Result[ValueT, ErrorT | ErrorT2]]: ...

    @overload
    def __call__(
        self, func: Callable[ParamT, Generator[Any, Any, ValueT]]
    ) -> Callable[ParamT, Result[ValueT, ErrorT]]: ...

    @overload
    def __call__(
        self, func: Callable[ParamT, Awaitable[Result[ValueT, ErrorT2]]]
    ) -> Callable[ParamT, Result[ValueT, ErrorT | ErrorT2]]: ...

    @overload
    def __call__(
        self, func: Callable[ParamT, Awaitable[ValueT]]
    ) -> Callable[ParamT, Result[ValueT, ErrorT]]: ...

    @overload
    def __call__(
        self, func: Callable[ParamT, Result[ValueT, ErrorT2]]
    ) -> Callable[ParamT, Result[ValueT, ErrorT | ErrorT2]]: ...

    @overload
    def __call__(
        self, func: Callable[ParamT, ValueT]
    ) -> Callable[ParamT, Result[ValueT, ErrorT]]: ...

    def __call__(
        self,
        func: Callable[ParamT, Generator[Any, Any, Result[ValueT, ErrorT2]]]
        | Callable[ParamT, Generator[Any, Any, ValueT]]
        | Callable[ParamT, Awaitable[Result[ValueT, ErrorT2]]]
        | Callable[ParamT, Awaitable[ValueT]]
        | Callable[ParamT, Result[ValueT, ErrorT2]]
        | Callable[ParamT, ValueT],
    ) -> (
        Callable[ParamT, Result[ValueT, ErrorT | ErrorT2]]
        | Callable[ParamT, Result[ValueT, ErrorT2]]
    ):
        @wraps(func)
        def inner(*args: ParamT.args, **kwargs: ParamT.kwargs) -> Result[ValueT, Any]:
            try:
                result = func(*args, **kwargs)
                if not isinstance(result, Container) and isinstance(result, Awaitable):
                    result = result.__await__()
                if isinstance(result, Generator):
                    result = unwrap_generator(result)
            except self._error_type as exc:
                return Error(exc)
            if isinstance(result, Error):
                return Error(result._other)  # noqa: SLF001
            if isinstance(result, Result):
                return Result(result._value)  # noqa: SLF001
            return Result(result)

        return inner


@overload
def catch() -> _Catch[Exception]: ...


@overload
def catch(*, error_type: type[ErrorT2]) -> _Catch[ErrorT2]: ...


@overload
def catch(*, error_type: type[Exception]) -> _Catch[Exception]: ...


@overload
def catch(func: None = ..., *, error_type: type[ErrorT2]) -> _Catch[ErrorT2]: ...


@overload
def catch(
    func: Callable[ParamT, Generator[Any, Any, Result[ValueT, ErrorT]]], /
) -> Callable[ParamT, Result[ValueT, ErrorT | Exception]]: ...


@overload
def catch(
    func: Callable[ParamT, Generator[Any, Any, ValueT]], /
) -> Callable[ParamT, Result[ValueT, Exception]]: ...


@overload
def catch(
    func: Callable[ParamT, Awaitable[Result[ValueT, ErrorT]]], /
) -> Callable[ParamT, Result[ValueT, ErrorT | Exception]]: ...


@overload
def catch(
    func: Callable[ParamT, Awaitable[ValueT]], /
) -> Callable[ParamT, Result[ValueT, Exception]]: ...


@overload
def catch(
    func: Callable[ParamT, Result[ValueT, ErrorT]], /
) -> Callable[ParamT, Result[ValueT, ErrorT | Exception]]: ...


@overload
def catch(
    func: Callable[ParamT, ValueT], /
) -> Callable[ParamT, Result[ValueT, Exception]]: ...


@overload
def catch(
    func: Callable[ParamT, Generator[Any, Any, Result[ValueT, ErrorT]]],
    /,
    *,
    error_type: type[ErrorT2],
) -> Callable[ParamT, Result[ValueT, ErrorT | ErrorT2]]: ...


@overload
def catch(
    func: Callable[ParamT, Generator[Any, Any, ValueT]], /, *, error_type: type[ErrorT2]
) -> Callable[ParamT, Result[ValueT, ErrorT2]]: ...


@overload
def catch(
    func: Callable[ParamT, Awaitable[Result[ValueT, ErrorT]]],
    /,
    *,
    error_type: type[ErrorT2],
) -> Callable[ParamT, Result[ValueT, ErrorT | ErrorT2]]: ...


@overload
def catch(
    func: Callable[ParamT, Awaitable[ValueT]], /, *, error_type: type[ErrorT2]
) -> Callable[ParamT, Result[ValueT, ErrorT2]]: ...


@overload
def catch(
    func: Callable[ParamT, Result[ValueT, ErrorT]], /, *, error_type: type[ErrorT2]
) -> Callable[ParamT, Result[ValueT, ErrorT | ErrorT2]]: ...


@overload
def catch(
    func: Callable[ParamT, ValueT], /, *, error_type: type[ErrorT2]
) -> Callable[ParamT, Result[ValueT, ErrorT2]]: ...


@overload
def catch(
    func: Callable[ParamT, Generator[Any, Any, Result[ValueT, ErrorT]]],
    /,
    *,
    error_type: type[Exception],
) -> Callable[ParamT, Result[ValueT, ErrorT | Exception]]: ...


@overload
def catch(
    func: Callable[ParamT, Generator[Any, Any, ValueT]],
    /,
    *,
    error_type: type[Exception],
) -> Callable[ParamT, Result[ValueT, Exception]]: ...


@overload
def catch(
    func: Callable[ParamT, Awaitable[Result[ValueT, ErrorT]]],
    /,
    *,
    error_type: type[Exception],
) -> Callable[ParamT, Result[ValueT, ErrorT | Exception]]: ...


@overload
def catch(
    func: Callable[ParamT, Awaitable[ValueT]], /, *, error_type: type[Exception]
) -> Callable[ParamT, Result[ValueT, Exception]]: ...


@overload
def catch(
    func: Callable[ParamT, Result[ValueT, ErrorT]], /, *, error_type: type[Exception]
) -> Callable[ParamT, Result[ValueT, ErrorT | Exception]]: ...


@overload
def catch(
    func: Callable[ParamT, ValueT], /, *, error_type: type[Exception]
) -> Callable[ParamT, Result[ValueT, Exception]]: ...


def catch(
    func: Callable[ParamT, Generator[Any, Any, Result[ValueT, ErrorT]]]
    | Callable[ParamT, Generator[Any, Any, ValueT]]
    | Callable[ParamT, Awaitable[Result[ValueT, ErrorT]]]
    | Callable[ParamT, Awaitable[ValueT]]
    | Callable[ParamT, Result[ValueT, ErrorT]]
    | Callable[ParamT, ValueT]
    | None = None,
    *,
    error_type: type[ErrorT2 | Exception] = Exception,
) -> (
    Callable[ParamT, Result[ValueT, ErrorT | Exception]]
    | Callable[ParamT, Result[ValueT, Exception]]
    | _Catch[ErrorT2]
    | _Catch[Exception]
):
    if func is None:
        return _Catch(error_type=error_type)

    @wraps(func)
    def inner(*args: ParamT.args, **kwargs: ParamT.kwargs) -> Result[ValueT, Any]:
        try:
            result = func(*args, **kwargs)
            if not isinstance(result, Container) and isinstance(result, Awaitable):
                result = result.__await__()
            if isinstance(result, Generator):
                result = unwrap_generator(result)
        except error_type as exc:
            return Error(exc)
        if isinstance(result, Error):
            return Error(result._other)  # noqa: SLF001
        if isinstance(result, Result):
            return Result(result._value)  # noqa: SLF001
        return Result(result)

    return inner



from __future__ import annotations

from typing import Any

from kontainer.container import Maybe, Result
from kontainer.core.const import undefined
from kontainer.decorator import catch, optional
from kontainer.flow import (
    bind_elements,
    compose_bind_funcs,
    compose_funcs,
    flip,
    flip_func,
    flowtools,
    map_elements,
    pipe_bind,
    pipe_iter_bind,
    pipe_iter_map,
    pipe_map,
    toggle,
)

__all__ = [
    "flowtools",
    "Maybe",
    "Result",
    "optional",
    "catch",
    "pipe_bind",
    "pipe_map",
    "pipe_iter_map",
    "pipe_iter_bind",
    "map_elements",
    "bind_elements",
    "compose_funcs",
    "compose_bind_funcs",
    "flip_func",
    "flip",
    "toggle",
    "undefined",
]

__version__: str


def __getattr__(name: str) -> Any:  # pragma: no cover
    from importlib.metadata import version

    if name == "__version__":
        return version("kontainer")

    error_msg = f"The attribute named {name!r} is undefined."
    raise AttributeError(error_msg)



from __future__ import annotations

from kontainer.flow import flowtools
from kontainer.flow.compose import compose_bind_funcs, compose_funcs
from kontainer.flow.convert import toggle
from kontainer.flow.flip import flip, flip_func
from kontainer.flow.maps import bind_elements, map_elements
from kontainer.flow.pipe import pipe_bind, pipe_iter_bind, pipe_iter_map, pipe_map

__all__ = [
    "pipe_map",
    "pipe_bind",
    "pipe_iter_map",
    "pipe_iter_bind",
    "map_elements",
    "bind_elements",
    "compose_funcs",
    "compose_bind_funcs",
    "flip_func",
    "flip",
    "flowtools",
    "toggle",
]
